// Christian Duncan
// Graph Class
//    for GD2004, Graph Drawing Contest

// package GD4;

import java.util.*;

/*********************
 * DenseGraph Class
 *    Represents an (Editable) Graph stored as an adjacency matrix
 *    Most ideal for dense graphs or speedy edge existence queries
 *********************/
public class DenseGraph extends Graph {
    private DenseNode node[];       // Info about all the Nodes
    private int nN;                 // Number of Nodes
    private int nE;                 // Number of Edges
    private boolean edge[][];       // The adjacency Matrix

    public DenseGraph(int n) {
	edge = new boolean[n][n];   // The whole matrix
	nN = nE = 0;                // None stored yet though
	node = new DenseNode[n];    // All the nodes
    }

    private DenseGraph() {
	edge = null; nN = nE = 0; node = null;
    }

    // Override the Graph clone method to make perfect clone
    public Graph deepClone() {
	DenseGraph copy = new DenseGraph();
	int i, j;
	int n = this.node.length;

	copy.edge = new boolean[n][n];
	copy.nN = this.nN;
	copy.nE = this.nE;
	copy.node = new DenseNode[n];

	// Copy the edges
	for (i = 0; i < n; i++)
	    for (j = 0; j < n; j++)
		copy.edge[i][j] = this.edge[i][j];

	// Copy the nodes
	for (i = 0; i < n; i++) {
	    Node thisNode = this.node[i];
	    copy.node[i] = new DenseNode(this.node[i], copy);
	}
	return copy;
    }

    // Add a node specifying its x and y locations
    public void addNode(double x, double y) {
	if (nN >= node.length) {
	    // Too many!
	    throw new RuntimeException("Programming Error: Too many nodes!");
	}
	node[nN] = new DenseNode(nN, x, y, this);
	nN++;
    }

    // Add an edge between two Nodes (referenced by index)
    //    Returns true if successful and false otherwise
    public boolean addEdge(int a, int b) {
	if (a < 0 || b < 0) {
	    // Too low
	    System.err.println("Error: Index too low!");
	    return false;
	} else if (a >= nN || b >= nN) {
	    // Too high
	    System.err.println("Error: Index too high!");
	    return false;
	}

	edge[a][b] = true;
	edge[b][a] = true;
	nE++;
	return true;
    }


    // Get list of all nodes (Iterator of Nodes)
    public Iterator nodeIterator() {
	return new DenseGraphNodeIterator(this);
    }

    // Get list of ALL edges (Iterator of Edges)
    public Iterator edgeIterator() {
	return new DenseGraphEdgeIterator(this);
    }

    public int numNodes() { return nN; }
    public int numEdges() { return nE; }

    /**********************************************************
     * This is a private class associated with DenseGraphs
     * DenseGraphNodeIterator -
     *     returns an iterator of the nodes in the given graph
     **********************************************************/
    private class DenseGraphNodeIterator implements Iterator {
	DenseGraph g;   // The current Graph to iterate on
	int n;          // Which node are we currently on

	public DenseGraphNodeIterator(DenseGraph _g) {
	    g = _g; n = 0;
	}

	public boolean hasNext() { return (n < g.numNodes()); }

	public Object next() {
	    if (n < g.numNodes()) {
		return node[n++]; // Return node n and THEN increment n
	    } 
	    else return null;     // No more elements!
	}

	public void remove() { /* Not implemented */ }
    }

    /**********************************************************
     * This is a private class associated with DenseGraphs
     * DenseGraphEdgeIterator -
     *     returns an iterator of the entire edges in the given graph
     * 
     *     Algorithm works as follows:
     *        nA,nB represents location in adj. matrix
     *                         between nodes nA and nB
     *        hasNext(): if mat[nA][nB] is an edge
     *                   then return true
     *                   otherwise find next edge (or set nA, nB to LARGE)
     *                              and return true/false
     *                   search only nB <= nA edges
     *                        This is the lower triangle
     *        next(): calls hasNext() first.
     *                if no - return null
     *                else nA,nB points to current one - then do advance()
     *        advance(): moves nB++ or nA++ (nB=0) and so on...
     **********************************************************/
    private class DenseGraphEdgeIterator implements Iterator {
	DenseGraph g;   // The current Graph to iterate on
	int nA;          // Which start node are we currently on
	int nB;          // Which end node are we currently on

	// Iterate edges over entire graph
	public DenseGraphEdgeIterator(DenseGraph _g) {
	    g = _g; nA = nB = 0;
	}

	public boolean hasNext() { 
	    // Skip until either end or found an edge
	    while (nA < g.numNodes() && g.edge[nA][nB] == false) {
		advance();  // Keep going until find empty edge
	    }
	    
	    return (nA < g.numNodes()); // Found edge or not
	}
	
	public Object next() {
	    if (hasNext() == false) return null;

	    // Create a new Edge object between the two nodes
	    Edge response = new Edge(g.node[nB], g.node[nA]);
	    advance();      // Skip this edge - hasNext advances next time
	    return response;
	}
	
	public void remove() { /* Not implemented */ }
	
	// Advance to next potential pair (in lower triangle)
	private void advance() {
	    nB++;
	    if (nB > nA) {
		// Passed lower triangle (next row)
		nA++;
		nB=0;
	    }
	}
    }

    /**********************************************************
     * This is a private class associated with DenseGraphs
     * DenseGraphNodeEdgeIterator -
     *     returns an iterator of the edges in the given graph's node
     * 
     *     Algorithm works as follows:
     *        nA,nB represents location in adj. matrix
     *                         between nodes nA and nB
     *        hasNext(): if mat[nA][nB] is an edge
     *                   then return true
     *                   otherwise find next edge (if exists)
     *                              and return true/false
     *        next(): calls hasNext() first.
     *                if no - return null
     *                else nA,nB points to current one - then do advance()
     **********************************************************/
    private class DenseGraphNodeEdgeIterator implements Iterator {
	DenseGraph g;   // The current Graph to iterate on
	int nA;          // Which start node are we using...
	int nB;          // Which end node are we currently on

	// Iterate edges over entire graph
	public DenseGraphNodeEdgeIterator(DenseGraph _g, int _node) {
	    g = _g; nA = _node; nB = 0;
	}

	public boolean hasNext() { 
	    // Skip until either end or found an edge
	    while (nB < g.numNodes() && g.edge[nA][nB] == false) {
		nB++;
	    }
	    
	    return (nB < g.numNodes()); // Found edge or not
	}
	
	public Object next() {
	    if (hasNext() == false) return null;

	    // Create a new Edge object between the two nodes
	    Edge response = new Edge(g.node[nA], g.node[nB]);
	    nB++;
	    return response;
	}
	
	public void remove() { /* Not implemented */ }
    }

    /*********************
     * A DenseNode in the Graph
     *    Stores:
     *       id - integer index into Graph Vector list (currently irrelevant!)
     *       X, Y - double location coordinates
     *       reference to the DenseGraph
     *********************/
    private class DenseNode extends Node {
	private DenseGraph g;

	public DenseNode(int _id, double _x, double _y, DenseGraph _g) {
	    super(_id, _x, _y);
	    g = _g;
	}

	public DenseNode(Node other, DenseGraph _g) {
	    super(other); g = _g;
	}

	/* Accessor methods */
	// Return iterator of incident edges
	public Iterator iterator() {
	    return new DenseGraphNodeEdgeIterator(g, id);
	}

	/* Add an edge to the list of edges
	 */
	void addEdge(Node other) {
	    // Create NEW Edge and add it to both
	    //    this and other edge lists
	    if (other instanceof DenseNode == false) {
		throw new RuntimeException("Error: DenseNode.addEdge(other): " +
					   "other MUST be instanceof DenseNode"); 
	    }
	    DenseNode o = (DenseNode) other;
	    if (this.g != o.g) {
		throw new RuntimeException("Error: DenseNode.addEdge(other): " +
					   "other MUST be in same graph as this node"); 
	    }		
	    g.addEdge(this.id, o.id);
	}
    }
}
