/*******************************
* GraphPanel:
*     Simple panel to draw a graph along with
*            its grid and other features...
*******************************/
// package GD4;

import javax.swing.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.event.*;
import java.awt.image.*;
import java.net.URL;
import java.util.*;

public class GraphPanel extends JPanel
implements Scrollable, MouseMotionListener, MouseListener {
    Graph graph;                              // The graph that we plan to draw
    double zoomLvl = 1.0;
    Rectangle graphBnds = null;
    Stack undoInfo;                           // Undo information
    final int MAX_UNDO = 20;                  // ... at least this much ...

    JToggleButton crossTrackerButton;
    int numCrossings = -1;

    // Colors, stroke patterns, and size of various objects in the display
    Color edgeColor = Color.black;            
    Color selectedEdgeColor = Color.gray;
    Color nodeColor = Color.red;              
    Color nodePinColor = new Color(0.5f, 0.0f, 0.0f); // A darker red
    Color crossColor = Color.blue;            
    Stroke edgeStroke = new BasicStroke((float) 4.0);
    Stroke selectedEdgeStroke = new BasicStroke((float) 3.0);
    Stroke nodeStroke = new BasicStroke((float) 2.0);
    Stroke crossStroke = new BasicStroke((float) 1.0);
    float nodeRadius = 10;
    float crossRadius = 7;

    final int DEFAULT_SCREEN_WIDTH = 1000;     //   Width/Height of Panel if no
    final int DEFAULT_SCREEN_HEIGHT = 1000;    //  Parent is provided
    final double D_ZOOM = 1.5;                // Factor to change zoom (per click)

   GraphPanel() {
      super (false);   // NOTE: Not sure why false - double-buffering should be preferred?
      this.graph = null;

      addMouseListener (this);
      addMouseMotionListener (this);

      // Let us set its size to a large value
      setBounds(0, 0, DEFAULT_SCREEN_WIDTH, DEFAULT_SCREEN_HEIGHT);

      undoInfo = new Stack();
   }

   /**
      * setGraph
    *    set the graph we are drawing
    *      calculate zoom level for this new graph
    **/
   public void setGraph(Graph g) {
      graph = g;
      fitToWindow();
   }

   /**
      * getGraph
    *    returns direct reference to Graph we are drawing
    **/
   public Graph getGraph() { return graph; }

    /**
     * setCrossTracker
     **/
    public void setCrossTracker(JToggleButton button) {
	crossTrackerButton = button;
    }

    /**
     * updateCrossTrackerTotal
     *    CTButton reflects total number of crossings (as it changes)
     *    recompute: true if we should recompute total crossings
     **/
    void updateCrossTrackerTotal(boolean recompute) {
	if (recompute && graph != null) {
	    // Recompute the number of intersections (crossings)
	    numCrossings = graph.countIntersections();
	}

	if (numCrossings >= 0) {
	    crossTrackerButton.setText(GraphMan.TK_CROSSTRACKER_ALT+": " +
				       numCrossings);
	} else {
	    crossTrackerButton.setText(GraphMan.TK_CROSSTRACKER_ALT);
	}
	crossTrackerButton.revalidate();
    }

   /**
    * calculateGraphBounds()
    *   Determine the best bounds for viewing the graph
    *   Relies on the Graph class to give bounds.
    *      May then add additional units for visual appeal
    **/
    void calculateGraphBounds() {
	// Compute bound of graph
	if (graph == null) return;

	graphBnds = graph.getBounds();
      
	// Add one block step in each direction
	int step = getBlockStep();
	graphBnds.x -= step;
	graphBnds.y -= step;
	graphBnds.width += 2*step;
	graphBnds.height += 2*step;
    }

   /**
      * fitToWindow()
    *     Calculate largest zoom value to fit graph
    *     onto the screen.
    **/
   void fitToWindow() {
       // Calculate screen width and height
       int screenWidth = DEFAULT_SCREEN_WIDTH;
       int screenHeight = DEFAULT_SCREEN_HEIGHT;

       if (getParent() != null) {
	   screenWidth = getParent().getWidth();
	   screenHeight = getParent().getHeight();
       }
       
       // Calculate graph width and height
       calculateGraphBounds();
       
       // Compute the scale in each direction
       double scaleX = (double) screenWidth / graphBnds.width;
       double scaleY = (double) screenHeight / graphBnds.height;
       
       // zoom level is smaller of two
       //    We don't want stretching
       zoomLvl = (scaleX < scaleY) ? scaleX : scaleY;
       
       revalidate();   // Force redraw if necessary
       repaint();
   }
    
    /**
      * zoomIn()
    *   Zoom the graph in
    *     Just change the zoomLevel and force redraw
    **/
   public void zoomIn() {
      zoomLvl *= D_ZOOM;    // Adjust stored zoom level

      // Force a redraw (NOTE: not sure how?)
      revalidate();
   }

   /**
      * zoomOut()
    *   Zoom the graph out
    *     Just change the zoomLevel and force redraw
    **/
   public void zoomOut() {
      zoomLvl /= D_ZOOM;

      // Force a redraw (NOTE: not sure how?)
      revalidate();
   }

   /**
      * setZoomLevel()
    *   Set the zoom level to the given value
    *     Just set the zoomLevel and force redraw
    **/
   public void setZoomLevel(double newZoomLvl) {
      if (newZoomLvl > 0) {
         // Valid value - set it
         zoomLvl = newZoomLvl;
         revalidate();
      }
   }

   /*********************
      * Methods needed for MouseMotion and Mouse Listeners
      *********************/
   public void mouseClicked (MouseEvent e) {
       if (e.getClickCount() == 1) {
	   /*
	   // Single-click, print Point Location for DEBUGING purposes
	   Point2D p = new Point(e.getX(), e.getY());
	   Point2D q = new Point();
	   computeAdjustedLocation(p, q);  // Where is point after scaling, translating?
	   Node nn = graph.nearestNode(q.getX(), q.getY(), nodeRadius);
	   if (nn == null) {
	   // Nothing selected
	   System.out.println("DEBUG: No point near: (" + 
	   q.getX() + "," + q.getY() + ")");
	   } else {
	   System.out.println("DEBUG: Selected node: " + 
	   nn.getId() + " (" + 
	   nn.getX() + "," + nn.getY() + ")");
	   }
	   */	   
	   return;
       }

       // Double-click indicates Pin/Unpin the selected node (if any)
       Point2D p = new Point(e.getX(), e.getY());
       Point2D q = new Point();
       computeAdjustedLocation(p, q);  // Where is point after scaling, translating?
      Node nn = graph.nearestNode(q.getX(), q.getY(), nodeRadius);
      if (nn == null) 
	  // Nothing selected
	  return;
	  
      nn.flipPinState();
      repaint();   // NOTE: Should only repaint AROUND the point actually!!
   }

   public void mouseEntered (MouseEvent e) {}
   public void mouseExited (MouseEvent e) {}
   public void mouseReleased (MouseEvent e) {}
   public void mouseMoved (MouseEvent e) {}

   /**
      * mousePressed:
    *     Identify the node (if any)
    *     that was selected
    **/
   public void mousePressed(MouseEvent e) {
      Point2D p = new Point(e.getX(), e.getY());
      Point2D q = new Point();
      computeAdjustedLocation(p, q);  // Where is point after scaling, translating?

      Node nn = graph.nearestNode(q.getX(), q.getY(), nodeRadius);

      // NOTE: Should do stuff with modifiers (multi-highlighted, etc.)
      if (nn != null) {
         graph.setSelectedNode(nn);
	 repaint();
      } else if (graph.hasSelectedNodes()) {
	  // Deselect
	  graph.unselectNodes();
	  repaint();
      }
   }

   /**
      * mouseDragged:
    *     If (moveable) node was selected move it
    *     to new location.
    **/
   public void mouseDragged(MouseEvent e) {
      Point2D p = new Point(e.getX(), e.getY());
      Point2D q = new Point();
      computeAdjustedLocation(p, q);  // Where is point after scaling, translating?

      graph.moveSelectedNodes(q.getX(), q.getY());
      calculateGraphBounds();

      revalidate();
      repaint();
   }

   /**
      * Convert the given located point
    *    to the adjust one - with scale and translation
    *    taken into account.
    *
    *    Given point is altered.
    **/
   public void computeAdjustedLocation(Point2D orig, Point2D adjusted) {
      // First compute the AffineTransformation
      AffineTransform aft = computeInverseAffineTransform();
      aft.transform(orig, adjusted);
   }

   /**
      * Compute proper affine transformation
    *   needed to adjust values from regular to
    *   graph's coordinate system - scale, translate, etc.
    **/
   public AffineTransform computeAffineTransform() {
      AffineTransform aft = new AffineTransform();
      aft.scale(zoomLvl, zoomLvl);
      
      if (graphBnds != null) {
	  // Translate so upper corner of bounding box
	  //   is at the origin (0,0)
	  aft.translate(-graphBnds.x, -graphBnds.y);
      }

      return aft;
   }

   /**
      * Compute proper (inverse) affine transformation
    *   needed to adjust values from regular to
    *   graph's coordinate system - scale, translate, etc.
    **/
   public AffineTransform computeInverseAffineTransform() {
      AffineTransform aft = new AffineTransform();

      if (graphBnds != null) {
	  // Translate so origin (0,0) goes to upper corner of 
	  //   bounding box of graph
	  aft.translate(graphBnds.x, graphBnds.y);
      }

      aft.scale(1.0/zoomLvl, 1.0/zoomLvl);
      return aft;
   }

   /*********************
      * Methods needed for Scrollable interface
      *********************/
   public Dimension getPreferredSize() {
      if (graph == null) {
         // No graph - no preferred size
	  return new Dimension(0, 0);
      }

      if (graphBnds == null)
	  calculateGraphBounds();

      // Based on dimensions of graph and zoom level
      int graphDX = (int) Math.round(graphBnds.width * zoomLvl);
      int graphDY = (int) Math.round(graphBnds.height * zoomLvl);

      return new Dimension(graphDX, graphDY);
   }

   public Dimension getPreferredScrollableViewportSize() {
      return getPreferredSize();
   }

   public int getScrollableUnitIncrement(Rectangle visibleRect,
                                         int orientation,
                                         int direction) {
      /**
      * This sample is for computing unit increments
       * for a picture.
       *  We need to do something different for our Graph
       *  Currently just returns 1
       *
       * //Get the current position.
       * int currentPosition = 0;
       * if (orientation == SwingConstants.HORIZONTAL) {
          * currentPosition = visibleRect.x;
          * } else {
             * currentPosition = visibleRect.y;
             * }
       *
       * //Return the number of pixels between currentPosition
       * //and the nearest tick mark in the indicated direction.
       * if (direction < 0) {
          * int newPosition = currentPosition -
          * (currentPosition / maxUnitIncrement)
          *                               * maxUnitIncrement;
          *             return (newPosition == 0) ? maxUnitIncrement : newPosition;
          * } else {
             * return ((currentPosition / maxUnitIncrement) + 1)
             * * maxUnitIncrement - currentPosition;
             * }
       **/
      return 10;
   }

   public int getScrollableBlockIncrement(Rectangle visibleRect,
                                          int orientation,
                                          int direction) {
      /** See comments from above... (for block increments)
      * if (orientation == SwingConstants.HORIZONTAL) {
         *  return visibleRect.width - maxUnitIncrement;
         * } else {
            *  return visibleRect.height - maxUnitIncrement;
            * }
      **/
      return getBlockStep();
   }

   public boolean getScrollableTracksViewportWidth() {
      return false;
   }

   public boolean getScrollableTracksViewportHeight() {
      return false;
   }

   /**
      * Using the graph and other info
    *   Compute the block step used for
    *   scrolling and for drawing grid space
    **/
   public int getBlockStep() {
      return 100;  // NOTE: Currently just hard-coded for 100
   }

    public void paintComponent (Graphics g) {
	// while (im == null)
	// im = (BufferedImage) createImage(getWidth()-20, getHeight()-20);
	// super.paintComponent(g);
	
	Graphics2D g2 = (Graphics2D) g;  // Use the more sophisticated 2D methods...
	int width, height;
	
	// Compute the affine transform to plot the points
	AffineTransform old = new AffineTransform(g2.getTransform()); // Save old one...
	AffineTransform aft = computeAffineTransform();
	g2.transform(aft);
	Rectangle visiblePart = g2.getClipBounds();
	
	// Clear ONLY the clipping area  (hopefully faster?)
	g2.setColor(getBackground());
	g2.fillRect(visiblePart.x, visiblePart.y, visiblePart.width, visiblePart.height);
	drawGrid(g2, visiblePart);
	drawGraph(g2, visiblePart);
	drawIntersections(g2, visiblePart);
	
	g2.setTransform(old);   // Restore original transform
    }

    /****
     * drawGrid
     *     Draws the grid but only the part within the
     *     given visible rectangle bounds - for SPEED!
     ****/
    void drawGrid(Graphics2D g, Rectangle visible) {
	int step = getBlockStep();
	int height = getHeight();
	int width = getWidth();
	
	// Need to calculate the first and last lines to draw
	int firstVert = (visible.x/step)*step;
	int lastVert = visible.x + visible.width;
	int firstHori = (visible.y/step)*step;
	int lastHori = visible.y + visible.height;
	
	g.setColor(Color.lightGray);
	
	// Draw vertical lines
	for (int i = firstVert; i <= lastVert; i+=step) {
	    g.drawLine(i, visible.y, i, lastHori);
	}
	
	// Draw horizontal lines
	for (int i = firstHori; i <= lastHori; i+= step) {
	    g.drawLine(visible.x, i, lastVert, i);
	}
    }
    
    /**
     * drawGraph
     *    Draws the current graph
     *      Ideally, only need to draw parts
     *      that hit visible region...
     *      But this can be somewhat tricky with
     *      edges... currently draws entire thing
     *      and lets clipping handle it
     **/
    void drawGraph (Graphics2D g, Rectangle visible) {
	// Is there a graph to draw?
	if (graph == null) return;
	
	Iterator it;
	
	// Draw all edges first
	it = graph.edgeIterator();
	while (it.hasNext()) {
	    Edge e = (Edge) it.next();
	    drawEdge(g, e);
	}
	
	// Draw all nodes now
	it = graph.nodeIterator();
	while (it.hasNext()) {
	    Node n = (Node) it.next();
	    drawNode(g, n);
	}
    }
    
    void drawEdge(Graphics2D g, Edge e) {
	// Get start and end nodes
	Node a = e.getStartNode();
	Node b = e.getEndNode();
	
	if (graph.isSelectedEdge(e)) {
	    g.setColor(selectedEdgeColor);
	    g.setStroke(selectedEdgeStroke);
	} else {
	    g.setColor(edgeColor);
	    g.setStroke(edgeStroke);
	}
	
	// For some annoying reason, graphics2D doesn't take double coordinates
	//    for line drawing!!!
	//    Will probably do some minimal scaling here so can handle small points!!!
	int ax = (int) Math.round(a.getX());
	int ay = (int) Math.round(a.getY());
	int bx = (int) Math.round(b.getX());
	int by = (int) Math.round(b.getY());
	
	g.drawLine(ax, ay, bx, by);
    }
    
    void drawNode(Graphics2D g, Node n) {
	// Distinguish between pinned and unpinned nodes
	if (n.isPinned()) {
	    g.setColor(nodePinColor);
	} else {
	    g.setColor(nodeColor);
	} 
	g.setStroke(nodeStroke);
	
	// Compute the starting x and y location
	double x = n.getX() - nodeRadius;
	double y = n.getY() - nodeRadius;
	
	// As with drawing an edge... round off values
	int cx = (int) Math.round(x);
	int cy = (int) Math.round(y);
	int rad = (int) Math.round(nodeRadius*2);
	
	g.fillOval(cx, cy, rad, rad);
	
	// If selected - highlight it as well
	if (graph.isSelectedNode(n)) {
	    g.setColor(selectedEdgeColor);
	    g.drawOval(cx, cy, rad, rad);
	}
    }
    
    /**
     * drawIntersections
     *    Draws all the intersections in the current graph
     **/
    void drawIntersections(Graphics2D g, Rectangle visible) {
	if (crossTrackerButton.isSelected() == false) {
	    // No crossings are required to be done
	    return;  // NOTE: Should probable empahsize crossing total (valid?)
	}
	
	// Is there even a graph to draw?
	if (graph == null) return;
	
	g.setColor(crossColor);
	g.setStroke(crossStroke);
	
	// Get the intersections
	Vector intersections = graph.findIntersections();
	Iterator it = intersections.iterator();
	
	while (it.hasNext()) {
	    Intersection i = (Intersection) it.next();
	    drawIntersection(g, i);
	}
	
	// Now, inform crossingTrackerButton of number of crossings
	numCrossings = intersections.size();
	updateCrossTrackerTotal(false);   // Don't need to recount crossings
    }
    
    void drawIntersection(Graphics2D g, Intersection inter) {
	// Compute the starting x and y location
	double x = inter.getX() - crossRadius;
	double y = inter.getY() - crossRadius;
	
	// As with drawing an edge... round off values
	int cx = (int) Math.round(x);
	int cy = (int) Math.round(y);
	int rad = (int) Math.round(crossRadius*2);
	
	g.fillOval(cx, cy, rad, rad);
    }
    
    /*****************************************************
     Routines for undo capability
        Includes full graph storage (for spring embedding)
        And individual movement
    *****************************************************/
    public void clearUndoInfo() { undoInfo.clear(); }

    public void checkSize() {
	if (undoInfo.size() > MAX_UNDO * 2) {
	    // Clear at all but the top MAX_UNDO pieces of information
	    Stack copy = new Stack();
	    int i;
	    for (i = 0; i < MAX_UNDO; i++)
		copy.push(undoInfo.pop());
	    undoInfo.clear();
	    for (i = 0; i < MAX_UNDO; i++)
		undoInfo.push(copy.pop());
	}
    }

    public void pushFullState() {
	if (graph == null) return;   // No graph to store!

	// Push the entire state onto the stack
	try {
	    Graph copy = graph.deepClone();  // Make a full copy
	    undoInfo.push(copy);
	    checkSize();
	} catch (CloneNotSupportedException e) {
	    // Hmm, can't clone this graph
	    // NOTE: We'll do nothing - if it can't save it, it can't save it.
	    return;
	}
    }

    public boolean undoEmpty() { return undoInfo.empty(); }

    /****
      Undo one level
      throws EmptyStackException if no more undo
    ****/
    public void undo() throws EmptyStackException {
	Object top = undoInfo.pop();
	
	if (top != null && top instanceof Graph) {
	    // Saved an entire graph
	    setGraph((Graph) top);
	} else {
	    System.err.println("Error: Unknown Undo Information");
	}
    } 
}
